<?php
/**
 * Plugin cisec
 * Copyright (c) Christophe IMBERTI
 * Licence Creative commons by-nc-sa
 */

// Si l'IP de l'internaute est bannie, afficher un message comprehensible
// avec indication du temps d'attente (utile en cas de faux positif)
function cisec_message($refresh_timeout, $redirection='./') {
    
    $texte = '<html>
    <head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
    <meta http-equiv="Content-Language" content="fr">
    <meta http-equiv="Refresh" content="'.$refresh_timeout.'; url='.$redirection.'">
    <title>Avertissement</title>
    <script>
    var count='.$refresh_timeout.';
    var counter=setInterval("timer()",1000); 
    function timer(){
      count=count-1;
      if (count <= 0){
         clearInterval(counter);
         return;
      }
     document.getElementById("timer").innerHTML=count;
    }
    </script>
    </head>
    <body style="font-family:Arial;font-size:2em;text-align:center;">
    <br>Votre demande est suspendue,
    <br>votre adresse IP a &eacute;t&eacute; enregistr&eacute;e.
    <br><br>Veuillez attendre : <span id="timer">'.$refresh_timeout.'</span> secondes<br>
    </body>
    </html>';

    // Pour les robots d'indexation en cas de faux positif
    // (meme methode que l'ecran de securite pour le 503)
    header("HTTP/1.1 503 Service Unavailable");
    header("Retry-After: ".$refresh_timeout); 
    header("Expires: Wed, 4 Jan 1984 00:00:00 GMT");
    header("Cache-Control: no-cache, must-revalidate");
    header("Pragma: no-cache");
    header("Content-Type: text/html");
    die($texte);
}

// Bannir temporairement une IP
// (un autre plugin peut appeler cette fonction s'il detecte une action suspecte)
function cisec_bannir($cause=''){
    $duree_banissement = _CISEC_DUREE_BANNISSEMENT;
    $trace_dir = cisec_trace_dir();
    $trace_nom = 'cisec_bannir';
    $trace_detail = 'cisec_detail';
    $temps = 0;
    $recidive = false;
    
    $fin_bannissement = cisec_est_banni();
    // si bannissement actif
    if ($fin_bannissement AND $fin_bannissement>time()){
        // ne pas enregistrer le bannissement a nouveau
        cisec_message($duree_banissement);

    } else {

        $temps = time() + $duree_banissement;
        
        // si recidive recente, allonger la duree de bannissement
        if ($fin_bannissement AND $fin_bannissement>time()-_CISEC_DUREE_BANNISSEMENT_RECIDIVE){
            $duree_banissement = _CISEC_DUREE_BANNISSEMENT_RECIDIVE;
            $temps = time() + $duree_banissement;
            $recidive = true;
        }
        
        // une virgule, une adresse IP, un pipe, un timestamp de fin de banissement, une virgule
        $texte = ','.cisec_ip().'_'.$temps.',';

        // enregistrer en version courte (pour de meilleures performances)
        cisec_enregistrer($texte, $trace_nom);

        // enregistrer en version detaillee (pour rechercher les causes plus tard)
	$texte_detail = date("Y-m-d H:i:s") . " | " . cisec_ip() . " (pid " . @getmypid() . ") | ".$cause." | "
		. "REQUEST_URI : " . $_SERVER["REQUEST_URI"] . "\n";
        cisec_enregistrer($texte_detail, $trace_detail);
        
        // envoyer un mail au webmestre
        // au plus un mail par tranche de 1 heure pour une meme adresse IP
        // pour eviter de saturer la boite aux lettre du webmestre
        if ($recidive){
            $destinataire = '';
            if (defined('_CISEC_EMAIL') AND _CISEC_EMAIL AND ctype_alnum(str_replace(array("@",".","_",'-'),'',_CISEC_EMAIL))){
                $destinataire = _CISEC_EMAIL;
            } elseif (isset($GLOBALS['meta']['email_webmaster']) AND $GLOBALS['meta']['email_webmaster']){
                $destinataire = $GLOBALS['meta']['email_webmaster'];
            }
            if ($destinataire){
                // un mail a-t-il deja ete envoye pour cette IP il y a moins de 1 heure ?
                $date_dernier_mail = cisec_date_dernier_mail();
                if (!$date_dernier_mail OR $date_dernier_mail<(time()-3600)){
                    // envoyer le mail
                    include_spip('inc/texte');
                    include_spip('inc/filtres');
                    // dans un fichier d'options la fonction typo de SPIP genere un warning sous PHP 7.4
                    $nom_site = textebrut($GLOBALS['meta']['nom_site']);
                    $adresse_site = textebrut($GLOBALS['meta']['adresse_site']);
                    $bannissement = "Bannissement temporaire de l'adresse IP ".cisec_ip();
                    $cause = textebrut($cause);
                    $raison = "pour la raison suivante : ".$cause;
                    $titre_mail = $nom_site . " : ".$bannissement." (".$cause.")";
                    $texte_mail = "Ceci est un message automatique du site : ".$nom_site." (".$adresse_site.") \r\n".$bannissement."\r\n".$raison;
                    $envoyer_mail = charger_fonction('envoyer_mail','inc');
                    $envoyer_mail($destinataire, $titre_mail, $texte_mail);

                    // enregistrer l'IP et la date d'envoi dans un log dedie a ces alertes par mail
                    cisec_enregistrer($texte, 'cisec_mail');
                }
            }
        }

        // afficher le message
        cisec_message($duree_banissement);
    }
}

// Enregistrer dans le log
function cisec_enregistrer($texte, $trace_nom){
    $trace_dir = cisec_trace_dir();
    $trace_suffix = '.log';
    $fichier = $trace_dir . $trace_nom . $trace_suffix;
    
    // verifier la taille
    $taille_trop_elevee = false;
    if ($GLOBALS['taille_des_logs']>0){
        if (@is_readable($fichier) AND @filesize($fichier) > $GLOBALS['taille_des_logs'] * 1024){
            $taille_trop_elevee = true;
        }
    }

    // ecrire
    $f = @fopen($fichier, 'ab');
    if ($f) {
            fputs($f, $texte);
            fclose($f);
    }
    
    // si taille trop elevee, rotation
    if ($taille_trop_elevee){
        $rotate = $GLOBALS['nombre_de_logs'];

        if ($rotate-- > 0 AND function_exists('spip_unlink')) {
            spip_unlink($fichier . '.' . $rotate);
            while ($rotate--) {
                    @rename($fichier . ($rotate ? '.' . $rotate : ''), $fichier . '.' . ($rotate + 1));
            }
        }
    }
}


// L'IP de l'internaute est elle bannie ?
// si oui, retourne le timestamp de fin de bannissement
function cisec_est_banni(){
    $return = false;
    $trace_dir = cisec_trace_dir();
    $fichier = $trace_dir . 'cisec_bannir.log';
    $recherche = ','.cisec_ip().'_';
    
    if ($contenu = @file_get_contents($fichier)){
        // bannissement le plus recent
        $debut = strrpos($contenu, $recherche);
        if ($debut!== false){
            $pos = $debut + strlen($recherche);
            $fin = strpos($contenu, ',', $pos);
            if ($fin!== false){
                $return = substr($contenu,$pos,$fin-$pos);
            }
        }
    }

    return $return;
}

// Detecter une signature de scanner dans l'URL ou les en-tetes HTTP
function cisec_signature_de_scanner(){
    $return = '';
    $signatures_variable_server = explode('|',_CISEC_SIGNATURES_VARIABLE_SERVER);
    $signatures_debut_user_agent = explode('|',_CISEC_SIGNATURES_DEBUT_USER_AGENT);
    $signatures_dans_url = explode('|',_CISEC_SIGNATURES_URL);
    $parties_url_suspectes = explode('|',_CISEC_PARTIES_URL_SUSPECTES);
    
    // Signatures de scanner comme variable _SERVER
    foreach ($signatures_variable_server as $signature){
        if (isset($_SERVER[$signature])) 
            $return = "Signature de scanner dans _SERVEUR : ".$signature;
    }
    
    // Signatures de scanner au debut du user agent
    if (!$return AND isset($_SERVER['HTTP_USER_AGENT'])){
        $ci_user_agent = strtolower($_SERVER['HTTP_USER_AGENT']);
        foreach ($signatures_debut_user_agent as $signature){
            if (substr($ci_user_agent,0,strlen($signature))==$signature)
                $return = "Signature de scanner dans HTTP_USER_AGENT : ".$signature;
        }
    }

    // Signatures de scanner dans l'URL
    $ci_request_uri = strtolower($_SERVER['REQUEST_URI']);
    if (!$return AND $ci_request_uri){
        foreach ($signatures_dans_url as $signature){
            if (strpos($ci_request_uri, $signature) !== false){
                $return = "Signature de scanner dans une URL : ".$signature;
                break;
            }
        }

        // la liste de ces mots peut etre enrichie via une constante dans un fichier d'options
        if (defined('_CISEC_MOTS')){
            if (_CISEC_MOTS){
                $cisec_mots = explode('|',_CISEC_MOTS);
                if ($cisec_mots AND is_array($cisec_mots)){
                    foreach ($cisec_mots as $cisec_mot){
                        if ($cisec_mot AND strpos($ci_request_uri, $cisec_mot) !== false){
                            // Ne pas propager le code malicieux
                            $return = "Mot interdit dans une URL (défini dans _CISEC_MOTS)";
                            break;
                        }
                    }
                }
            }
        }
    }
    
    
    // tentative d'injection sur forcer_lang
    // (l'ecran de securite filtre lang mais pas forcer_lang)
    if (!$return AND isset($_GET['forcer_lang'])){
        if ($_GET['forcer_lang'] AND $_GET['forcer_lang']!="true" AND $_GET['forcer_lang']!="false")
            $return = "Tentative d'injection sur forcer_lang";
    }

    // tentative d'injection sur lang
    // (attention : l'ecran de securite filtre lang en amont
    // et met un espace en remplacement d'un caractere interdit)
    if (!$return AND isset($_GET['lang'])){
        if ($_GET['lang'] AND !ctype_alpha(str_replace('_','',$_GET['lang'])))
            $return = "Tentative d'injection sur lang";
        elseif ($_GET['lang'] AND strpos(trim($_GET['lang']), ' ') !== false)
            $return = "Tentative d'injection sur lang";
    }
    
    // tentative d'injection sur la pagination
    if (!$return AND strpos($ci_request_uri, 'debut_')!== false){
        foreach ($_GET as $var => $valeur){
            if (substr($var,0,6)=='debut_'){
                if (!is_numeric($valeur))
                    $return = "Tentative d'injection sur la pagination";
            } 
        }
    }
  
    // Partie d'URL suspecte
    if (!$return AND $ci_request_uri){
        foreach ($parties_url_suspectes as $signature){
            if (strpos($ci_request_uri, $signature) !== false){
                $return = "Partie d'URL suspecte ".$signature;
                break;
            }
        }
    }
    
    return $return;
}

// Enregistrer un log a chaque POST
// et indiquer si les seuils sont depasses
// ou si POST suspect
function cisec_flood_post(){
    $return = '';
    $trace_nom = 'cisec_post';

    // Uniquement en cas de POST
    // Espace prive et CISF doivent toujours rester en dehors du champ
    if (!empty($_POST) AND strpos($_SERVER['REQUEST_URI'], "ecrire/") === false 
            AND strpos($_SERVER['REQUEST_URI'], "/spip.php?page=cisf_") === false){
        
        // ne pas comptabiliser l'AJAX de sauvegarde en session (formulaire FORUM de SPIP)
        if (isset($_POST['action']) AND $_POST['action']=='session')
            return $return;

        // POST suspect dans un formulaire
        if ( (isset($_POST['nobot']) AND $_POST['nobot'])
                OR (isset($_POST['email_nobot']) AND $_POST['email_nobot']) ){
            $return = "POST d'un champ leurre dans un formulaire";
        
        } else {    
        // Les seuils sont-ils depasses (anti flood) ?
            $trace_dir = cisec_trace_dir();
            $trace_suffix = '.log';
            $fichier = $trace_dir . $trace_nom . $trace_suffix;

            // flood ?
            $time = time();
            $recherche = ','.cisec_ip().'_';
            $contenu = @file_get_contents($fichier);

            if ($contenu){
                // prendre les 600 derniers caracteres
                // soit environ les 20 dernieres traces
                $index = 600;
                $longueur = strlen($contenu);
                if ($longueur<$index)
                    $index = $longueur;
                $contenu = substr($contenu,-$index);

                $seuil_court = 3;
                $duree_court = 1;           
                $seuil_long = 9;
                $duree_long = 10;
                
                if (defined('_CISEC_MAX_POST_EN_1_S') AND intval(_CISEC_MAX_POST_EN_1_S)>$seuil_court){
                        $seuil_court = intval(_CISEC_MAX_POST_EN_1_S);
                }
                if (defined('_CISEC_MAX_POST_EN_10_S') AND intval(_CISEC_MAX_POST_EN_10_S)>$seuil_long){
                        $seuil_long = intval(_CISEC_MAX_POST_EN_10_S);
                }

                $nombre_post = substr_count($contenu, $recherche);

                // cette IP a deja envoye plusieurs POST
                if ($nombre_post>=$seuil_court){
                    $timstamp_court = $time - $duree_court + 1;
                    $timstamp_long = $time - $duree_long + 1;
                    $nb_court = 0;
                    $nb_long = 0;

                    preg_match_all('#'.$recherche.'(.*?),#', $contenu, $regs);

                    foreach ($regs[1] as $timestamp) {
                        if ($timestamp>=$timstamp_court)
                            $nb_court++;
                        if ($timestamp>=$timstamp_long)
                            $nb_long++;
                    }

                    if ($nb_court>=$seuil_court)
                        $return = $nb_court." POST de cette IP en ".$duree_court." seconde";
                    elseif ($nb_long>=$seuil_long)
                        $return = $nb_long." POST de cette IP en ".$duree_long." secondes";
                }
            }
        }
        
        // enregistrer
        $texte = ','.cisec_ip().'_'.$time.',';
        cisec_enregistrer($texte, $trace_nom);
    }    
    return $return;
}

// Date du dernier mail envoye pour cette IP
function cisec_date_dernier_mail(){
    $return = false;
    $trace_dir = cisec_trace_dir();
    $fichier = $trace_dir . 'cisec_mail.log';
    
    $recherche = ','.cisec_ip().'_';
    
    if ($contenu = @file_get_contents($fichier)){
        // mail le plus recent pour cette IP
        $debut = strrpos($contenu, $recherche);
        if ($debut!== false){
            $pos = $debut + strlen($recherche);
            $fin = strpos($contenu, ',', $pos);
            if ($fin!== false){
                $return = substr($contenu,$pos,$fin-$pos);
            }
        }
    }
    return $return;
}

function cisec_trace_dir(){
    static $dir = '';
    if (!$dir){
	$dir = ((defined('_DIR_LOG') AND !defined('_DIR_PLUGIN_CIMS')) ? _DIR_LOG : _DIR_RACINE._NOM_TEMPORAIRES_INACCESSIBLES);
        if (defined('_DIR_PLUGIN_CIMS') AND function_exists($f_racine = 'cims_racine_dossier')){
            $dir = $f_racine(_NOM_TEMPORAIRES_INACCESSIBLES)._NOM_TEMPORAIRES_INACCESSIBLES;
        }
    }
    return $dir;
}

// Pour ceux qui utilisent un reverse proxy
// et qui ne definissent pas la valeur de $ip dans mes_options.php
// (et qui n'utilisent pas mod_rpaf ou mod_remoteip)
function cisec_ip() {
    static $return = '';
    
    if (!$return){
	// par defaut
	$return = $GLOBALS['ip'];
	
	// ordre de recherche par defaut
	$ordre = array('HTTP_X_FORWARDED_FOR','REMOTE_ADDR');
	
	// ordre de recherche personnalise dans la constante _CI_ORDRE_RECHERCHE_IP
	if (defined('_CI_ORDRE_RECHERCHE_IP') AND _CI_ORDRE_RECHERCHE_IP)
		$ordre = explode(",",_CI_ORDRE_RECHERCHE_IP);

	// determination de l'IP
	foreach ($ordre as $valeur) {
                $valeur = trim($valeur);
		if (isset($_SERVER[$valeur]) AND $_SERVER[$valeur]) {
			$return = $_SERVER[$valeur];
			break;
		}
	}
    }
    
    return $return;
}


// Duree du bannissement (en secondes)
define("_CISEC_DUREE_BANNISSEMENT", 30); 
// Duree du bannissement en cas de recidive (en secondes)
define("_CISEC_DUREE_BANNISSEMENT_RECIDIVE", 180); 
// Signatures de scanner comme variable _SERVER
define("_CISEC_SIGNATURES_VARIABLE_SERVER", "HTTP_ACUNETIX_PRODUCT|HTTP_ACUNETIX_SCANNING_AGREEMENT|HTTP_ACUNETIX_USER_AGREEMENT");
// Signatures de scanner au debut du user agent
define("_CISEC_SIGNATURES_DEBUT_USER_AGENT", "arachni|useragent");
// Signatures de scanner dans l'URL
define("_CISEC_SIGNATURES_URL", "acunetix_wvs_security_test|_arachni_trainer_|vvv0000|"
        ."injected_by_wvs|tests.arachni-scanner.com");
// Partie d'URL suspecte
define("_CISEC_PARTIES_URL_SUSPECTES", "../../../../|.././.././|page=recherche&page=recherche");

// Espace prive et CISF doivent rester en dehors du champ
if (strpos($_SERVER['REQUEST_URI'], "ecrire/") === false 
        AND strpos($_SERVER['REQUEST_URI'], "/spip.php?page=cisf_") === false){
    $fin_bannissement = cisec_est_banni();
    // si bannissement actif
    if ($fin_bannissement AND $fin_bannissement>time())
        cisec_message($fin_bannissement-time());
    elseif($cause = cisec_signature_de_scanner())
        cisec_bannir($cause);
    elseif ($cause = cisec_flood_post())
        cisec_bannir($cause);
}

?>